# -*- coding: utf-8 -*-

import random
import copy, sys
import numpy as np
from string import Template

def joueur_humain(etat, fct_but, fct_transitions, str_joueur):
    
    actions_str = '; '.join([ a.__str__() for a in fct_transitions(etat).keys()])
    action = raw_input('Entrer un choix d\'action puis appuyer sur Enter.\nChoisir parmi: {'+actions_str+'}\n')
    while True:
        try:
            exec 'action_ret = ' + action
            fct_transitions(etat)[action_ret]
            break
        except:
            action = raw_input('L\'action n\'est pas valide. Réessayer à nouveau, puis appuyer sur Enter\n')
    return action_ret
    
def joueur_aleatoire(etat, fct_but, fct_transitions, str_joueur):
    action,etat = random.choice(fct_transitions(etat).items())
    return action


class Jeu:
    """
    Classe de jeu.
    
    Initialize à partir de la fonction but,
    fonction transitions et l'état inital du jeu.
    
    La méthode "jouer_partie" simule une partie 
    où un joueur Max (les 'X') et un joueur Min (les 'O')
    s'affrontent.
    """
    def __init__(self, etat_initial, fct_but, fct_transitions):
        self.but = fct_but
        self.transitions = fct_transitions
        self.etat_initial = etat_initial

    def print_message(self,resultat):
        if resultat > 0:
            print 'Joueur X a gagné'
        elif resultat == 0:
            print 'Partie nulle'
        else:
            print 'Joueur O a gagné'

    def jouer_partie(self,joueur_max, joueur_min):
        etat = copy.deepcopy(self.etat_initial)
        print etat
        while True:
            action = joueur_max(copy.deepcopy(etat),self.but,self.transitions,'X')
            etat = self.transitions(etat)[action]
            print etat
            resultat = self.but(etat)
            if resultat is not None:
                self.print_message(resultat)
                break
            action = joueur_min(copy.deepcopy(etat),self.but,self.transitions,'O')
            etat = self.transitions(etat)[action]
            print etat
            resultat = self.but(etat)
            if resultat is not None:
                self.print_message(resultat)
                break


#############################################
# Etat, transitions et but pour le Connect4 #
#############################################

class Connect4Etat:

    def __init__(self):
        self.n_colonnes = 8
        self.n_rangees = 6
        self.tableau = np.array([[' ']*self.n_colonnes]*self.n_rangees)

    def __str__(self):
        ret = ''
        for i,r in enumerate(self.tableau):
            ret += '|'+'|'.join(r)+'|\n'
        ret += '-'*(2*self.n_colonnes+1)+'\n'
        ret += ' '+' '.join([str(i) for i in range(self.n_colonnes)])+'\n'
        return ret

def connect4_transitions(etat):
    # Determiner c'est le tour à qui
    if np.sum(etat.tableau == 'X') > np.sum(etat.tableau == 'O'):
        # Tour de O
        symbol = 'O'
    else:
        # Tour de X
        symbol = 'X'

    actions = {}
    for i in range(etat.n_colonnes):
        try:
            rangee_pour_i = np.nonzero(etat.tableau.T[i] == ' ')[0].max()
            nouvel_etat = copy.deepcopy(etat)
            nouvel_etat.tableau[rangee_pour_i,i] = symbol
            actions[i] = nouvel_etat
        except:
            pass
    return actions

def connect4_but(etat):
    def symbol_gagne(symbol):
        gagne = False
        #for i in range(etat.n_rangees):
        #    for j in range(etat.n_colonnes):
        if np.sum(etat.tableau==symbol) == 0:
            return False
        x,y = np.nonzero(etat.tableau==symbol)
        for i,j in zip(x,y):
                # Vérifie 4 pareil sur une rangée...
                gagne |= np.sum(etat.tableau[i,j:j+4] == symbol) == 4
                # ... sur une colonne
                gagne |= np.sum(etat.tableau[i:i+4,j] == symbol) == 4
                # ... sur une diagonale gauche-droite
                gagne |= np.sum(np.diag(etat.tableau[i:i+4,j:j+4]) == symbol) == 4
                # ... sur une diagonale droite-gauche
                gagne |= np.sum(np.diag(etat.tableau[i-4+1:i+1,j:j+4][::-1]) == symbol) == 4
                if gagne:
                    return gagne
        return gagne

    # Vérifie si X a gagné
    X_gagne = symbol_gagne('X')
    if X_gagne:
        #return np.inf
        return 100000+np.sum(etat.tableau == ' ')

    # Vérifie si O a gagné
    O_gagne = symbol_gagne('O')    
    if O_gagne:
        #return -np.inf
        return -100000-np.sum(etat.tableau == ' ')

    # Vérifie si c'est une partie nulle
    if np.sum(etat.tableau != ' ') == etat.n_colonnes*etat.n_rangees:
        return 0
    return None

    
################################
# Execution en tant que script #
################################

def main():
    usage = """Usage: python connect4.py adversaire

où "adversaire" est "humain", "soi-meme" ou "aleatoire"."""

    if len(sys.argv) != 2:
        print usage
        return None
    else:
        adversaire = sys.argv[1]
        
    if adversaire not in ['aleatoire','humain','soi-meme']:
        print usage
        return None

    
    # Jouer une partie de Connect4
    connect4 = Jeu(Connect4Etat(), connect4_but, connect4_transitions)
                                 
    import solution_connect4 as solution
    
    int_tempsMaximal = 10
    def joueur(etat,fct_but,fct_transitions,str_joueur):
        return solution.joueur_connect4(etat,fct_but,fct_transitions,str_joueur,int_tempsMaximal)

    if adversaire == 'aleatoire':
        # Pour jouer contre un joueur aleatoire
        connect4.jouer_partie(joueur,joueur_aleatoire)
    
    if adversaire == 'humain':
        # Pour jouer contre un joueur humain
        connect4.jouer_partie(joueur,joueur_humain)
    
    if adversaire == 'soi-meme':
        # Pour joueur contre sa propre solution
        connect4.jouer_partie(joueur,joueur)

if __name__ == "__main__":
    main()
