# -*- coding: utf-8 -*-

import random
import copy, sys
import numpy as np
from string import Template

def joueur_humain(etat, fct_but, fct_transitions, str_joueur):
    
    actions_str = '; '.join([ a.__str__() for a in fct_transitions(etat).keys()])
    action = raw_input('Entrer un choix d\'action puis appuyer sur Enter.\nChoisir parmi: {'+actions_str+'}\n')
    while True:
        try:
            exec 'action_ret = ' + action
            fct_transitions(etat)[action_ret]
            break
        except:
            action = raw_input('L\'action n\'est pas valide. Réessayer à nouveau, puis appuyer sur Enter\n')
    return action_ret
    
def joueur_aleatoire(etat, fct_but, fct_transitions, str_joueur):
    action,etat = random.choice(fct_transitions(etat).items())
    return action


class Jeu:
    """
    Classe de jeu.
    
    Initialize à partir de la fonction but,
    fonction transitions et l'état inital du jeu.
    
    La méthode "jouer_partie" simule une partie 
    où un joueur Max (les 'X') et un joueur Min (les 'O')
    s'affrontent.
    """
    def __init__(self, etat_initial, fct_but, fct_transitions):
        self.but = fct_but
        self.transitions = fct_transitions
        self.etat_initial = etat_initial

    def print_message(self,resultat):
        if resultat > 0:
            print 'Joueur X a gagné'
        elif resultat == 0:
            print 'Partie nulle'
        else:
            print 'Joueur O a gagné'

    def jouer_partie(self,joueur_max, joueur_min):
        etat = copy.deepcopy(self.etat_initial)
        print etat
        while True:
            action = joueur_max(copy.deepcopy(etat),self.but,self.transitions,'X')
            etat = self.transitions(etat)[action]
            print etat
            resultat = self.but(etat)
            if resultat is not None:
                self.print_message(resultat)
                break
            action = joueur_min(copy.deepcopy(etat),self.but,self.transitions,'O')
            etat = self.transitions(etat)[action]
            print etat
            resultat = self.but(etat)
            if resultat is not None:
                self.print_message(resultat)
                break


################################################
# Etat, transitions et but pour le tic-tac-toe #
################################################

class TicTacToeEtat:

    def __init__(self):
        self.tableau = np.array([ [' ',' ',' '],
                                  [' ',' ',' '],
                                  [' ',' ',' '] ])

    def __str__(self):
        t = Template("""
   0   1   2
     |   |   
0  $a | $b | $c
     |   |   
  ---+---+---
     |   |   
1  $d | $e | $f
     |   |
  ---+---+---
     |   |   
2  $g | $h | $i
     |   |
""")
        return t.substitute(a=self.tableau[0,0],b=self.tableau[0,1],c=self.tableau[0,2],
                            d=self.tableau[1,0],e=self.tableau[1,1],f=self.tableau[1,2],
                            g=self.tableau[2,0],h=self.tableau[2,1],i=self.tableau[2,2])

def tictactoe_transitions(etat):
    # Determiner c'est le tour à qui
    if np.sum(etat.tableau == 'X') > np.sum(etat.tableau == 'O'):
        # Tour de O
        symbol = 'O'
    else:
        # Tour de X
        symbol = 'X'
    positions_vides = np.nonzero(etat.tableau == ' ')
    actions = {}
    for i,j in zip(positions_vides[0],positions_vides[1]):
        nouvel_etat = copy.deepcopy(etat)
        nouvel_etat.tableau[i,j] = symbol
        actions[(i,j)] = nouvel_etat
    return actions

def tictactoe_but(etat):
    # Vérifie si X a gagné
    X_gagne = False
    X_gagne |= etat.tableau[0,0] == 'X' and etat.tableau[0,1] == 'X' and etat.tableau[0,2] == 'X' 
    X_gagne |= etat.tableau[1,0] == 'X' and etat.tableau[1,1] == 'X' and etat.tableau[1,2] == 'X' 
    X_gagne |= etat.tableau[2,0] == 'X' and etat.tableau[2,1] == 'X' and etat.tableau[2,2] == 'X' 
    X_gagne |= etat.tableau[0,0] == 'X' and etat.tableau[1,0] == 'X' and etat.tableau[2,0] == 'X' 
    X_gagne |= etat.tableau[0,1] == 'X' and etat.tableau[1,1] == 'X' and etat.tableau[2,1] == 'X' 
    X_gagne |= etat.tableau[0,2] == 'X' and etat.tableau[1,2] == 'X' and etat.tableau[2,2] == 'X' 
    X_gagne |= etat.tableau[0,0] == 'X' and etat.tableau[1,1] == 'X' and etat.tableau[2,2] == 'X' 
    X_gagne |= etat.tableau[0,2] == 'X' and etat.tableau[1,1] == 'X' and etat.tableau[2,0] == 'X' 
    if X_gagne:
        return 100000+np.sum(etat.tableau == ' ')

    # Vérifie si O a gagné
    O_gagne = False
    O_gagne |= etat.tableau[0,0] == 'O' and etat.tableau[0,1] == 'O' and etat.tableau[0,2] == 'O' 
    O_gagne |= etat.tableau[1,0] == 'O' and etat.tableau[1,1] == 'O' and etat.tableau[1,2] == 'O' 
    O_gagne |= etat.tableau[2,0] == 'O' and etat.tableau[2,1] == 'O' and etat.tableau[2,2] == 'O' 
    O_gagne |= etat.tableau[0,0] == 'O' and etat.tableau[1,0] == 'O' and etat.tableau[2,0] == 'O' 
    O_gagne |= etat.tableau[0,1] == 'O' and etat.tableau[1,1] == 'O' and etat.tableau[2,1] == 'O' 
    O_gagne |= etat.tableau[0,2] == 'O' and etat.tableau[1,2] == 'O' and etat.tableau[2,2] == 'O' 
    O_gagne |= etat.tableau[0,0] == 'O' and etat.tableau[1,1] == 'O' and etat.tableau[2,2] == 'O' 
    O_gagne |= etat.tableau[0,2] == 'O' and etat.tableau[1,1] == 'O' and etat.tableau[2,0] == 'O' 
    if O_gagne:
        return -100000-np.sum(etat.tableau == ' ')

    # Vérifie si c'est une partie nulle
    if np.sum(etat.tableau != ' ') == 9:
        return 0
    return None

    
################################
# Execution en tant que script #
################################

def main():
    usage = """Usage: python tictactoe.py adversaire

où "adversaire" est "humain", "soi-meme" ou "aleatoire"."""

    if len(sys.argv) != 2:
        print usage
        return None
    else:
        adversaire = sys.argv[1]
        
    if adversaire not in ['aleatoire','humain','soi-meme']:
        print usage
        return None

    # Jouer une partie de Tic-tac-toe
    tictactoe = Jeu(TicTacToeEtat(), tictactoe_but, tictactoe_transitions)
    
    import solution_tictactoe as solution
    
    if adversaire == 'aleatoire':
        # Pour jouer contre un joueur aleatoire
        tictactoe.jouer_partie(solution.joueur_tictactoe,joueur_aleatoire)
    
    if adversaire == 'humain':
        # Pour jouer contre un joueur humain
        tictactoe.jouer_partie(solution.joueur_tictactoe,joueur_humain)
    
    if adversaire == 'soi-meme':
        # Pour joueur contre sa propre solution
        tictactoe.jouer_partie(solution.joueur_tictactoe,solution.joueur_tictactoe)


if __name__ == "__main__":
    main()
